
# assert = function(expr)
#   if(!eval(expr, parent.frame(n = 6)))
#     stop("Assertion failed for expression ", paste(expr))
assert = function(expr)
  if(!expr)
    stop("Assertion failed")

.enharmonicSpellings        <- c("-B",  "a", "+c", "-E",  "f",  "E",  "e",  "+f", "-A", "B",  "G",   "-A")
names(.enharmonicSpellings) <- c("+A", "--b", "-d", "+D", "+e", "-F", "-f", "-g", "+G", "-c", "--A", "--B")

#######################################################################################
## readKSFileAsTable() reads a KSN file and converts it into a table format
readKSFileAsTable = function(filename, enharmonic=TRUE, ignoreloops=FALSE)
{
  KSData = readKSFile(filename, ignoreloops=ignoreloops)
  tableData = matrix(NA, nrow(KSData), 3+2+15)
  colnames(tableData) = c("Measures", "Beats", "Ticks", names(interpretKSKey(KSData[1, 2], enharmonic)), names(interpretKSChord(KSData[1, 3])))
  lengths = KSData[,1]
  lengths = cbind(lengths / KSData[,5], lengths)
  lengths = cbind(lengths[,1] / KSData[,4], lengths)
  tableData[, 1:3] = lengths
  for(ii in 1:nrow(tableData))
    tableData[ii, 4:ncol(tableData)] = c(interpretKSKey(KSData[ii, 2], enharmonic), interpretKSChord(KSData[ii, 3]))
  return(tableData)
}

#######################################################################################
## saveKSAsTable() saves a harmony table to a text file
saveKSAsTable = function(KSTable, filename)
  write.table(KSTable, filename, row.names=FALSE)

#######################################################################################
## readKSfile() loads a KSN file and returns a column matrix with 5 columns:
##   Ticks  - length of the chord in ticks
##   Key    - key
##   Chord  - chord
##   BPM    - number of beats per measure
##   Beat   - number of eight notes per beat, multiplied by 3 (number of ticks per beat)
##
## A tick is always a third of an eight note, so there are 96 ticks per whole note.
## To get the number of ticks per measure:
##     ticksPerMeasure = beatsPerMeasure * beatValue
## To get the meter:
##     meter = beatsPerMeasure \ (96/beatValue)
##
readKSFile = function(filename, ignoreloops=FALSE)
{
  .keyModifiers <- matrix(0, 30, 7)
  rownames(.keyModifiers) <- c("C", "D", "E", "F", "G", "A", "B", "c", "d", "e", "f", "g", "a", "b", "+C", "+d", "+F", "+f", "+g", "+a", "+c", "-C", "-D", "-E", "-e", "-G", "-A", "-a", "-B", "-b")
  .keyModifiers[1,]  <-  1*c(0, 0, 0, 0, 0, 0, 0)  # C-major
  .keyModifiers[2,]  <-  1*c(0, 0, 1, 0, 0, 0, 1)  # D-major
  .keyModifiers[3,]  <-  1*c(0, 1, 1, 0, 0, 1, 1)  # E-major
  .keyModifiers[4,]  <- -1*c(0, 0, 1, 0, 0, 0, 0)  # F-major
  .keyModifiers[5,]  <-  1*c(0, 0, 0, 0, 0, 0, 1)  # G-major
  .keyModifiers[6,]  <-  1*c(0, 0, 1, 0, 0, 1, 1)  # A-major
  .keyModifiers[7,]  <-  1*c(0, 1, 1, 0, 1, 1, 1)  # B-major
  .keyModifiers[8,]  <- -1*c(0, 0, 1, 0, 0, 1, 1)  # c-minor
  .keyModifiers[9,]  <- -1*c(0, 0, 0, 0, 0, 1, 0)  # d-minor
  .keyModifiers[10,] <-  1*c(0, 1, 0, 0, 0, 0, 0)  # e-minor
  .keyModifiers[11,] <- -1*c(0, 0, 1, 1, 0, 1, 1)  # f-minor
  .keyModifiers[12,] <- -1*c(0, 0, 1, 0, 0, 1, 0)  # g-minor
  .keyModifiers[13,] <- -1*c(0, 0, 0, 0, 0, 0, 0)  # a-minor
  .keyModifiers[14,] <-  1*c(0, 1, 0, 0, 1, 0, 0)  # b-minor
  .keyModifiers[15,] <-  1*c(1, 1, 1, 1, 1, 1, 1)  # C#-major
  .keyModifiers[16,] <-  1*c(1, 1, 1, 1, 1, 0, 1)  # d#-minor
  .keyModifiers[17,] <-  1*c(1, 1, 1, 0, 1, 1, 1)  # F#-major
  .keyModifiers[18,] <-  1*c(1, 1, 0, 0, 1, 0, 0)  # f#-minor
  .keyModifiers[19,] <-  1*c(1, 1, 0, 1, 1, 0, 1)  # g#-minor
  .keyModifiers[20,] <-  1*c(1, 1, 1, 1, 1, 1, 1)  # a#-minor
  .keyModifiers[21,] <-  1*c(1, 1, 0, 1, 1, 0, 0)  # c#-minor
  .keyModifiers[22,] <- -1*c(1, 1, 1, 1, 1, 1, 1)  # Cb-major
  .keyModifiers[23,] <- -1*c(1, 1, 0, 1, 1, 1, 0)  # Db-major
  .keyModifiers[24,] <- -1*c(1, 0, 0, 1, 1, 0, 0)  # Eb-major
  .keyModifiers[25,] <- -1*c(1, 0, 1, 1, 1, 1, 1)  # eb-minor
  .keyModifiers[26,] <- -1*c(1, 1, 1, 1, 1, 1, 0)  # Gb-major
  .keyModifiers[27,] <- -1*c(1, 1, 0, 1, 1, 0, 0)  # Ab-major
  .keyModifiers[28,] <- -1*c(1, 1, 1, 1, 1, 1, 1)  # ab-minor
  .keyModifiers[29,] <- -1*c(1, 0, 0, 1, 0, 0, 0)  # Bb-major
  .keyModifiers[30,] <- -1*c(1, 0, 1, 1, 0, 1, 1)  # bb-minor

  .specialTokens = c("|", "||", ":|", "|:", ":||", "||:")

  ############################################
  ## Internal functions

  checkBraces = function(tokens, leftbrace="{", rightbrace="}", stoponerror=TRUE)
  {
    if(leftbrace == "{")
      lbracepos = grep("\\{", tokens)
    else if(leftbrace == "[")
      lbracepos = grep("\\[", tokens)
    else if(leftbrace == "(")
      lbracepos = grep("\\(", tokens)

    if(rightbrace == "}")
      rbracepos = grep("\\}", tokens)
    else if(rightbrace == "]")
      rbracepos = grep("\\]", tokens)
    else if(rightbrace == ")")
      rbracepos = grep("\\)", tokens)

    bracePositions = c(lbracepos, rbracepos)
    if(length(bracePositions) == 0)
      return(0)

    openBrackets = c(rep(1, length(lbracepos)), rep(-1, length(rbracepos)))
    sortPositions = sort(bracePositions, index.return=TRUE)
    openBrackets = openBrackets[sortPositions$ix]
    bracePositions = sortPositions$x
    openBrackets = cumsum(openBrackets)
    if(any(openBrackets < 0) && stoponerror)
    {
      errorPos = bracePositions[which(openBrackets < 0)[1]]
      if(errorPos < length(tokens))
        stop("Unexpected closing brace at: ", tokens[errorPos])
      else
        stop("Unexpected closing brace at: ", tokens[errorPos])
    }
    if((openBrackets[length(openBrackets)] != 0) && stoponerror)
      stop(paste(openBrackets[length(openBrackets)], "unmatched brackets!"))
    return(openBrackets[length(openBrackets)])
  }

  getRelativeKey = function(key, relative, isminor)
  {
    relative = sub("=[\\+-ABCDEFGabcdefg]+$", "", relative)   # remove everything after equation mark
    relative = sub("^[/0123456789]+", "", relative)
    if(grepl("^[\\+-]*[ABCDEFGabcdefg]$", relative))
      return(relative)
    keyNames = c("A", "B", "C", "D", "E", "F", "G")
    if(grepl("^[\\+-]", relative))
    {
      modifier = substr(relative, 1, 1)
      relative = substr(relative, 2, nchar(relative))
    }
    else
      modifier = NULL

    keyLetter = toupper(substr(key, nchar(key), nchar(key)))
    relativeNumber = as.integer(as.roman(relative))
    relativeKey    = keyNames[(which(keyNames == keyLetter) + relativeNumber - 2) %% 7 + 1]
    if(isminor)
      relativeKey = tolower(relativeKey)
    if(!(key %in% rownames(.keyModifiers)))
      key = .enharmonicSpellings[key]
    if(.keyModifiers[key, relativeNumber] > 0)
      relativeKey = paste("+", relativeKey, sep="")
    if(.keyModifiers[key, relativeNumber] < 0)
      relativeKey = paste("-", relativeKey, sep="")
    if(!is.null(modifier))
    {
      if(modifier == "+")
        relativeKey = paste("+", relativeKey, sep="")
      else if(modifier == "-")
        relativeKey = paste("-", relativeKey, sep="")
    }
    relativeKey = sub("\\+-", "", relativeKey)
    relativeKey = sub("-\\+", "", relativeKey)
    return(relativeKey)
  }

  compressKSsequence = function(ksSequence)
  {
    lengths   = rle(paste(ksSequence[1,], ksSequence[2,], ksSequence[3,], ksSequence[4,]))$lengths
    positions = c(1, 1+cumsum(lengths[-length(lengths)]))
    return(data.frame(Ticks=lengths, Key=ksSequence[1, positions], Chord=ksSequence[2,positions], BPM=as.numeric(ksSequence[3,positions]), Beat=as.numeric(ksSequence[4,positions]), stringsAsFactors=FALSE))
  }

  barContainsModulation = function(tokens)
  {
    result = rep(FALSE, length(tokens))
    for(ii in 1:length(tokens))
      if(regexpr("\\{[abcdefgABCDEFG\\+-iIvV]+\\:$", tokens[ii]) != -1)
        result[ii] = TRUE
    return(result)
  }

  splitLengthChord = function(chord)  # separates chord length from the rest of the chord definition
  {
    chordregexp  = "(^[0-9./]*)(.+$)"
    notevalue    = sub(chordregexp, "\\1", chord)
    chord        = sub(chordregexp, "\\2", chord)

    if(notevalue == "")
      return(list(chordlen=1, chord=chord))
    else
      return(list(chordlen=notevalue, chord=chord))
  }

  isTokenChord = function(token) (regexpr("^[iIvI:+-]*[0-9/]*[abcdefgABCDEFGivIV!+-]+", token) != -1)

  modulationsLeftOpen = function(tokens)
  {
    modulationOpens  = which(barContainsModulation(tokens))
    modulationCloses = which(tokens == "}")
    for(ii in modulationCloses)
      modulationOpens[rev(which((modulationOpens < ii) & (modulationOpens != -1)))[1]] = -1
    return(tokens[modulationOpens[modulationOpens != -1]])
  }

  expandLoop = function(repBeg, repEnd, tokens)
  {
    if(is.na(repEnd) || (tokens[repEnd] == ":||"))  # normal repetition
    {
      if(is.na(repEnd))
      {
	repEnd = length(tokens) + 1
	beforeRep = c(tokens[1:(repBeg-1)], "|")
	afterRep = c()
	leftOpen = modulationsLeftOpen(beforeRep)
      }
      else if(repEnd == length(tokens))
      {
	beforeRep = c(tokens[1:(repBeg-1)], "|")
	afterRep = c()
	leftOpen = modulationsLeftOpen(beforeRep)
      }
      else if(is.na(repBeg) || (repEnd < repBeg))
      {
	repBeg = 0
	beforeRep = c()
	afterRep = c(tokens[(repEnd+1):length(tokens)])
	leftOpen = c()
      }
      else
      {
	beforeRep = c(tokens[1:(repBeg-1)], "|")
	afterRep = c(tokens[(repEnd+1):length(tokens)])
	if((length(afterRep) == 1) && (afterRep == ":||"))
	  afterRep = c()
	leftOpen = modulationsLeftOpen(beforeRep)
      }
      repetition = c(tokens[(repBeg+1):(repEnd-1)], "|")
      openBraces = rep("}", checkBraces(tokens[1:(repEnd-1)], "{", "}", FALSE))
      return(c(beforeRep, repetition, openBraces, leftOpen, repetition, afterRep))
    }
    else if(tokens[repEnd] == ":|[2")  # repetition with alternative endings
    {
      if((repBeg < 1) | (repEnd > length(tokens)))
        stop("Unsupported repetition")
      firstAltBeg  = which(tokens == "|[1")
      firstAltBeg  = firstAltBeg[which((firstAltBeg > repBeg) & (firstAltBeg < repEnd))[1]]
      beforeRep    = c(tokens[1:(repBeg-1)], "|")
      repeatedPart = tokens[(repBeg+1):(firstAltBeg-1)]
      firstAlt     = tokens[(firstAltBeg+1):(repEnd-1)]
      secondAlt    = tokens[(repEnd+1):length(tokens)]
      beforeImbalance = checkBraces(beforeRep, "{", "}", FALSE)
      beforeOpen      = modulationsLeftOpen(tokens[1:(repBeg-1)])
      return(c(beforeRep, repeatedPart, "|", firstAlt, rep("}", beforeImbalance), beforeOpen, "|", repeatedPart, "|", secondAlt))
    }
    else
      stop("Internal error: bad repetition end mark")
  }

  ############################################
  ## Read the input file line by line
  lines = readLines(filename)
  tokens = c()  # list of individual tokens from the KSN file
  for(ii in 1:length(lines)) # parse the input file
  {
    curline = sub("%.*$", "", lines[ii]) # remove the comment
    curline = sub(' +$', '', curline)    # remove trailing spaces
    curline = sub('^ +', '', curline)    # remove initial spaces
    curline = sub('\\s+', ' ', curline)  # change all white space to single spaces
    tokens = c(tokens, unlist(strsplit(curline, "\\s+", perl=TRUE)))  # split into tokens on all empty characters
  }

  ############################################
  ## Pre-process the token list

  # Contract redundant chord definitions that contain spaces
  # (e.g. +vi-7-''=[III VI+ I' V']) into single tokens
  newtokens = c()
  insideChord = FALSE
  for(ii in 1:length(tokens))
  {
    if(grepl("^.+=\\[.+", tokens[ii]))
    {
      insideChord = TRUE
      newtokens = c(newtokens, tokens[ii])
    }
    else if(insideChord)
    {
      if(grepl(".+\\]$", tokens[ii]))
        insideChord = FALSE
      newtokens[length(newtokens)] = paste(newtokens[length(newtokens)], tokens[ii])
    }
    else
      newtokens = c(newtokens, tokens[ii])
  }
  tokens = newtokens
  rm(newtokens)

  # Remove entries like " V}={V: I " (e.g. pivot chords).
  # Such a sequence will be converted to " } {V: I ", so only the second 
  # chord interpretation will be used.
  newtokens = c()
  for(ii in 1:length(tokens))
  {
    if(grepl("^[^@]+=[^\\[]+", tokens[ii]))
    {
      cat("Simplifying redundant (e.g. pivot) chord definition", tokens[ii])
      closingBraces = length(unlist(strsplit(tokens[ii], "}", fixed=TRUE))) - 1
      simplifiedChord = c(rep("}", closingBraces), unlist(strsplit(tokens[ii], "=", fixed=TRUE))[2])
      newtokens = c(newtokens, simplifiedChord)
      cat(" to", simplifiedChord, "\n")
    }
    else
      newtokens = c(newtokens, tokens[ii])
  }
  tokens = newtokens
  rm(newtokens)

  # Expand multi-token square brackets (e.g. [2C E 2G])
  # (Note: they now seem to be banned from the standard)
#   openSquareBrackets = 0
#   for(ii in 1:length(tokens))
#   {
#     bracketBalance = checkBraces(tokens[ii], "[", "]", stoponerror=FALSE)
#     if(bracketBalance > 0)
#     {
#       openSquareBrackets = openSquareBrackets + 1
#       tokens[ii] = paste(tokens[ii], "]", sep="")
#     }
#     else if(bracketBalance < 0)
#     {
#       openSquareBrackets = openSquareBrackets - 1
#       tokens[ii] = paste("[", tokens[ii], sep="")
#     }
#     else if((openSquareBrackets > 0) && !(tokens[ii] %in% .specialTokens))
#       tokens[ii] = paste("[", tokens[ii], "]", sep="")
#   }

  # Expand multi-token pedal points (e.g. "[V]&{!V9 | i'}")
  pedalPoint = ""
  for(ii in 1:length(tokens))
  {
    if(regexpr("\\[.+\\]\\&\\{.+", tokens[ii]) != -1)  # a pedal point starts here
    {
      splitToken = unlist(strsplit(tokens[ii], "\\&\\{"))
      pedalPoint = splitToken[1]
      pedalPoint = paste(splitLengthChord(unlist(strsplit(pedalPoint, "[\\[\\]]", perl=TRUE))[2])$chord, sep="")
         # ignore the length of the pedal point
      tokens[ii] = paste(splitToken[2], "/", pedalPoint, sep="")  # add pedal point at the end
      cat("Copying pedal point", pedalPoint, "\n")
    }
    else if((pedalPoint != "") && (regexpr(".*\\}$", tokens[ii]) != -1))  # a pedal point ends here
    {
      if(nchar(tokens[ii]) > 1)  # more than closing brace
        tokens[ii] = paste(sub("\\}", "", tokens[ii]), "/", pedalPoint, sep="")  # add pedal point at the end
      else
        tokens[ii] = ""
      pedalPoint = ""
    }
    else if((pedalPoint != "") && !(tokens[ii] %in% .specialTokens))
      tokens[ii] = paste(tokens[ii], "/", pedalPoint, sep="")  # add pedal point at the end
  }
  tokens = tokens[tokens != ""]  # remove empty tokens left after removing braces that were closing pedal points

  # Pull the note values to the beginning of each token
  # and add "no-harmony" ("q") to all explicit note lists
  for(ii in 1:length(tokens))
  {
    if(regexpr("^\\[.+\\]$", tokens[ii]) != -1) # no note value before brackets
    {
      tokens[ii] = unlist(strsplit(tokens[ii], "[\\[\\]]", perl=TRUE))[2]  # remove the brackets
      splitChord = splitLengthChord(tokens[ii])
      if(splitChord$chordlen == "1")
        tokens[ii] = paste("q&[", splitChord$chord, "]", sep="")
      else
        tokens[ii] = paste(splitChord$chordlen, "q&[", splitChord$chord, "]", sep="")
    }
    else if(regexpr("^[0-9/]+\\[.+\\]$", tokens[ii]) != -1)  # note value at the beginning
    {
      splitChord = unlist(strsplit(tokens[ii], "[\\[\\]]", perl=TRUE))
      tokens[ii] = paste(splitChord[1], "q&[", splitChord[2], "]", sep="")
    }
  }

  # Remove all chords in parenthesis (I assume they only contain _alternative_ interpretation)
  for(ii in 1:length(tokens))
  {
    if(regexpr("^\\(.+\\)", tokens[ii]) != -1)
    {
      cat("Simplifying redundant chord definition", tokens[ii])
      if(regexpr(".+\\}$", tokens[ii]) != -1)
        closingBrace = "}"
      else        
        closingBrace = ""
      tokens[ii] = sub("\\}$", "", tokens[ii])  # remove the closing brace
      currentChord  = splitLengthChord(unlist(strsplit(tokens[ii], "[\\(\\)]", perl=TRUE))[2])
      if(!isTokenChord(tokens[ii-1])) # is previous token a chord or other special symbol
      {
        if(ii-2 < 1)
          stop("Cannot simplify redundant chord (not enough chords before it)!")
        previousChord = splitLengthChord(tokens[ii-2])
        keyChangeToken = tokens[ii-1]
        tokens[ii-1] = paste(currentChord$chordlen, previousChord$chord, closingBrace, sep="")  # use previous chord with current note value
        tokens[ii] = keyChangeToken
        cat(" to", tokens[ii-1], "\n")
      }
      else
      {
        previousChord = splitLengthChord(tokens[ii-1])
        tokens[ii] = paste(currentChord$chordlen, previousChord$chord, closingBrace, sep="")  # use previous chord with current note value
        cat(" to", tokens[ii], "\n")
      }
    }
  }
  tokens = tokens[tokens != ""]  # remove empty tokens

  # Split multiple closing braces and brackets joint with chord definitions
  newtokens = c()
  for(ii in 1:length(tokens))
  {
    if((bracesCount = attributes(regexpr("\\}+$", tokens[ii]))$match.length) > -1)
    {
      whatsleft = sub("\\}+$", "", tokens[ii])
      if(nchar(whatsleft) > 0)
        newtokens = c(newtokens, whatsleft, rep("}", bracesCount))
      else
        newtokens = c(newtokens, rep("}", bracesCount))
    }
    else
      newtokens = c(newtokens, tokens[ii])
  }
  tokens = newtokens
  rm(newtokens)

  # Replace repetition marks with repeated chords
  for(ii in 1:length(tokens))
  {
    if(grepl("_", tokens[ii]))
    {
      splitChord = unlist(strsplit(tokens[ii], "_"))
      if(length(splitChord) == 1)  # only "_"
      {
        leftPart = splitChord
        rightPart = ""
      }
      else
      {
        leftPart = splitChord[1]
        rightPart = splitChord[2]
      }
      for(jj in (ii-1):1)
      {
        if(isTokenChord(tokens[jj]))
        {
          middlePart = splitLengthChord(tokens[jj])$chord
          break
        }
      }
      tokens[ii] = paste(leftPart, middlePart, rightPart, sep="")
    }
  }

  # Expand higher-level loops (da capo al segno, etc.)
  while(any(tokens %in% c("@DCAF", "@DCAC", "@DSAF", "@DSAC")))
  {
    DCAF = which(tokens == "@DCAF")
    if(length(DCAF) > 1)
      stop("There can be only one @DCAF in a song")
    else if(length(DCAF) > 0)
    {
      fine = which(tokens == "@F")
      if(length(fine) > 1)
        stop("There can be only one @F in a song")
      else if(length(fine) > 0)
      {
        imbalance = checkBraces(tokens[1:(fine-1)], "{", "}", FALSE)
        tokens = c(tokens[1:(fine-1)], tokens[(fine+1):(DCAF-1)], "|", tokens[1:(fine-1)], rep("}", imbalance), "||")
      }
      else  #TODO:Unchecked!
      {
        imbalance = checkBraces(tokens[1:(DCAF-1)], "{", "}", FALSE)
        tokens = c(tokens[1:(DCAF-1)], rep("}", imbalance), tokens[1:(DCAF-1)], tokens[(DCAF+1):length(tokens)], "||")
      }
    }
  }

  # Expand simple loops
  tokens[tokens == ":|"] = ":||"
  tokens[tokens == "|:"] = "||:"
  tokens[tokens == ":|:"] = ":||:"
  while(any(tokens == ":||:"))  # convert :|: to :|| ||:
  {
    loopPos = which(tokens == ":||:")[1]
    tokens = c(tokens[1:(loopPos-1)], ":||", "||:", tokens[(loopPos+1):length(tokens)])
  }
  if(!ignoreloops)
  {
    while(any((tokens == ":||") | (tokens == "||:")))  # expand all the loops
    {
      repBeg = which(tokens == "||:")[1]
      repEnd = which((tokens == ":||") | (tokens == ":|[2"))[1]
      tokens = expandLoop(repBeg, repEnd, tokens)
    }
  }
  else
  {
    tokens[tokens == ":||"] = "|"
    tokens[tokens == "||:"] = "|"
  }

  # Sanity check
  if(checkBraces(tokens, "{", "}") != 0)
    stop("Unbalanced brackets!")

  # Measure mark cleanup
  tokens[tokens == "||"] = "|"      # replace "||"s with "|"s
  tokens = tokens[nchar(tokens) > 0]  # remove empty elements
  tokens = tokens[(c(FALSE, tokens[2:length(tokens)] != tokens[1:(length(tokens)-1)]) | (tokens != "|"))] # Remove repeated "|" marks

  ##########################################33
  ## Token analysis
  bpmSequence   = c()
  bValSequence  = c()
  chordSequence = c()
  keySequence   = c()
  curBPM   = NA
  curBVal  = NA
  curBar   = c()
  curKeys  = c()
  keyStack = c()
  for(ii in 1:length(tokens))  # process the data
  {
    curkey = keyStack[length(keyStack)]
    if((regpos = regexpr("^@M", tokens[ii])) != -1)
    {
      metrum = substr(tokens[ii], 4, nchar(tokens[ii]))
      curBPM = as.integer(sub("/.+", "", metrum))   # beats per measure
      curBVal = 96 / as.integer(sub(".+/", "", metrum))  # beat length in sixteenths
      next
    }
    if((regpos = regexpr("^@K", tokens[ii])) != -1)
    {
      keyStack = substr(tokens[ii], 4, nchar(tokens[ii]))
      curkey = keyStack[length(keyStack)]
      next
    }
    if((regpos = regexpr("^@", tokens[ii])) != -1)
    {
      stop("Unsupported directive ", tokens[ii])
      next
    }
    if((regpos = regexpr("^\\{[abcdefgABCDEFG\\+-iIvV]+\\:$", tokens[ii])) != -1)  # modulation start
    {
      newKey = substr(tokens[ii], 2, nchar(tokens[ii]) - 1)
      newKey = getRelativeKey(keyStack[length(keyStack)], newKey, newKey == tolower(newKey))
      keyStack = c(keyStack, newKey)
      curkey = newKey
      next
    }
    if((regpos = regexpr("\\}$", tokens[ii])) != -1)
    {
      keyStack = keyStack[-length(keyStack)]
      curchord = substr(tokens[ii], 1, nchar(tokens[ii]) - 1)
    }
    else
      curchord = tokens[ii]
    if(curchord == "")   # can it happen?
      next

    if(curchord != "|")
    {
      if((regexpr("^@", curchord) == -1) || (regexpr("^\\[[123456789]+$", curchord) == -1))  # ignore directives
      {
        curBar   = c(curBar, curchord)
        curKeys  = c(curKeys, curkey)
      }
    }
    else if(!is.null(curBar))
    {
      lengths = rep(0,  length(curBar))
      chords  = rep("", length(curBar))
      for(jj in 1:length(curBar))
      {
        splitchord = splitLengthChord(curBar[jj])
        if(grepl("/", splitchord$chordlen))
        {
          fraclen = strsplit(splitchord$chordlen, "/")[[1]]
          lengths[jj] = as.numeric(fraclen[1])/as.numeric(fraclen[2])
        }
        else
          lengths[jj] = as.numeric(splitchord$chordlen)
        chords[jj] = splitchord$chord
      }
      if(sum(lengths) == 0)
        lengths = 1 / length(lengths)
      else
        lengths = lengths / sum(lengths)
      lengths = lengths * curBPM * curBVal
      keys = rep("", length(chords))
      for(jj in 1:length(keys))
      {
        if(grepl(":", chords[jj]))
        {
          keychord = strsplit(chords[jj], ":")[[1]]
          keys[jj] = getRelativeKey(curKeys[jj], keychord[1], keychord[1] == tolower(keychord[1]))
          chords[jj] = keychord[2]
        }
        else
          keys[jj] = curKeys[jj]
      }
      for(jj in 1:length(chords))
        chords[jj] = sub("=.*$", "", chords[jj])  # remove everything after equation mark
      chordSequence = c(chordSequence, rep(chords, times=lengths))
      keySequence   = c(keySequence, rep(keys, times=lengths))
      bpmSequence   = c(bpmSequence, rep(curBPM, times=sum(lengths)))
      bValSequence  = c(bValSequence, rep(curBVal, times=sum(lengths)))
      if(length(chordSequence) != length(bpmSequence))
        stop("Looks like the tick is too small or there is a note value error in this measure: ", paste(curBar, collapse=" "))
      curBar = c()
      curKeys = c()
    }
  }

  newKeySequence = rep("", length(keySequence))
  curKey = keySequence[1]
  newKeySequence[1] = curKey
  for(ii in 2:length(keySequence))
  {
    if(keySequence[ii] != keySequence[ii-1])
      curKey = keySequence[ii]
    newKeySequence[ii] = curKey
  }

  return(compressKSsequence(rbind(keys=newKeySequence, chords=chordSequence, beatsPerMeasure=bpmSequence, beatValue=bValSequence)))
}

######################################################################################################
## interpretKSKey()
##
## Parameters:
##   - key
##     A character string with key information
##   - enahrmonic
##     A Boolean indicating what format do we want the data to be returned in. If TRUE (default value),
##     it will return the number of accidentals(?) (-7...0...+7), where negative numbers indicate
##     flats and positive numbers sharps, and a mode. If FALSE, it will return the absolute position
##     of the tonic (0..11), where 0 ic C, and a mode. Mode is zero for major keys and 1 for minor
##     ones.
##
interpretKSKey = function(key, enharmonic=TRUE)
{
  if(key %in% names(.enharmonicSpellings))
    key = .enharmonicSpellings[key]

  tonicPositions = c("C"=0, "D"=2, "E"=4, "F"=5, "G"=7, "A"=9, "B"=11)
  if(toupper(key) == key)
    mode = 0
  else
    mode = 1

  splitRegexp = "([\\+-]*)([abcdefgABCDEFG]*)"
  accidental = sub(splitRegexp, "\\1", key)
  tonic      = toupper(sub(splitRegexp, "\\2", key))
  if((nchar(tonic) != 1) || is.na(tonicPositions[tonic]))
    stop("Invalid key:", key)

  if(!enharmonic)
  {
    tonic = tonicPositions[tonic]
    if(accidental == "+")
      tonic = tonic + 1
    else if(accidental == "-")
      tonic = tonic - 1
    else if(accidental == "++")
      tonic = tonic + 2
    else if(accidental == "--")
      tonic = tonic - 2
    tonic = tonic %% 12
    names(tonic) = c()
    return(c(Tonic=tonic, Mode=mode))
  }
  else
  {
    if(mode == 0)
      Signature = which(c("-C", "-G", "-D", "-A", "-E", "-B", "F", "C", "G", "D", "A",  "E",  "B",  "+F", "+C") == key) - 8
    else
      Signature = which(c("-a", "-e", "-b", "f",  "c",  "g",  "d", "a", "e", "b", "+f", "+c", "+g", "+d", "+a") == key) - 8
    return(c(Signature=Signature, Mode=mode))
  }
}

######################################################################################################
## - scale degree (1..7, 0 for non-chord)
## - chord type (0 = major, 1 = minor)
## - inversion (0,1,...)
## - member note modifiers: (NA, 0, +1, -1, +2, -2)
##   - root modifier
##   - 2nd modifier
##   - 3rd modifier
##   - 4th modifier
##   - 5th modifier
##   - 6th modifier
##   - 7th modifier
##   - 9th modifier
##   - 11th modifier
##   - 13th modifier
##   - 15th modifier
##   - added note
##
##  For note modifier: NA means that pitch is not present, 0 means that it is present.
##  Other numbers correspond to raising or lowering the pitch by the given number of semitones
##
interpretKSChord = function(chord, tonic=NA, mode=NA)
{
  pitchNamesSharp = c("C", "+C", "D", "+D", "E", "F", "+F", "G", "+G", "A", "+A", "B")
  pitchNamesFlat  = c("C", "-D", "D", "-E", "E", "F", "-G", "G", "-A", "A", "-B", "B")
  majorDegrees    = which(c(1,0,1,0,1,1,0,1,0,1,0,1) == 1) - 1  # a major scale
  minorDegrees    = which(c(1,0,1,1,0,1,0,1,1,0,1,0) == 1) - 1  # a harmonic minor scale

  chordData = c(Degree=NA, Type=NA, Inversion=NA, Root=NA, Second=NA, Third=NA, Fourth=NA, Fifth=NA, Sixth=NA, Seventh=NA, Ninth=NA, Eleventh=NA, Thirteenth=NA, Fifteenth=NA, Added=NA, Pedal=NA)

  originalChord = chord

  # use single characters for double flats and sharps (simpler to process)
  chord = sub("\\+\\+", "#", chord)
  chord = sub("--", "=", chord)

  chordRegexp = "^([!#=\\+-]*)([^!1-9\\=#&'/\\+-]+)([^'&/]*)('*)&?([^/]*)/?([A-Ga-gVIvi\\+-]*)$"
  rootModifiers  = sub(chordRegexp, "\\1", chord)
  degree         = sub(chordRegexp, "\\2", chord)
  otherModifiers = sub(chordRegexp, "\\3", chord)
  inversion      = sub(chordRegexp, "\\4", chord)
  addedNotes     = sub(chordRegexp, "\\5", chord)
  pedalNote      = sub(chordRegexp, "\\6", chord)

  inversion = nchar(inversion)

  if(grepl("^\\[.+\\]$", degree))     # explicit member notes
  {
    chordData["Degree"] = 0
    degree = toupper(unlist(strsplit(sub("^\\[(.+)\\]$", "\\1", degree), " "))[1])  # only the first member note is taken
    if(grepl("[a-gA-G+\\-#=]+", degree))      # absolute pitch 
      chordData["Added"] = NA #TODO
    else if(grepl("[1-7+\\-#=]+", degree))    # pitch relative to root
      chordData["Added"] = NA #TODO
    else if(grepl("[IViv+\\-#=]+", degree))   # pitch relative to tonic
    {
      if(is.na(tonic))
        stop("Pitch relative to tonic is specified, but no tonic is given in ", originalChord)
      chordData["Added"] = NA # TODO
    }
    else
      stop("Unrecognized explicit member note declaration in ", originalChord)
  }
  else
  {
    if(degree %in% c("z", "q"))
      chordData["Degree"] = 0
    else
    {
      chordData["Inversion"] = inversion
      chordData["Degree"]    = as.integer(as.roman(degree))
      chordData["Root"]      = 0
      chordData["Third"]     = 0
      chordData["Fifth"]     = 0
      if(toupper(degree) == degree)  # major chord
	chordData["Type"] = 0
      else if(tolower(degree) == degree)  # minor chord
	chordData["Type"] = 1
      else
	stop("Chord degree has mixed capital and small letters: ", chord)

      if(nchar(rootModifiers) > 1)
	stop("Too many root modifiers in ", chord, "!")
      if(rootModifiers == "!")
	chordData["Root"] = NA
      else if(rootModifiers == "+")
	chordData["Root"] = 1
      else if(rootModifiers == "-")
	chordData["Root"] = -1
      else if(rootModifiers == "#")
	chordData["Root"] = +2
      else if(rootModifiers == "=")
	chordData["Root"] = -2

      fifthModifiers = sub("([^0-9]*)([0-9]*.*)", "\\1", otherModifiers)
      otherModifiers = sub("([^0-9]*)([0-9]*.*)", "\\2", otherModifiers)
      if(nchar(fifthModifiers) > 1)
	stop("Too many fifth modifiers in ", chord, "!")
      if(fifthModifiers == "!")
	chordData["Fifth"] = NA
      else if(fifthModifiers == "+")
	chordData["Fifth"] = 1
      else if(fifthModifiers == "-")
	chordData["Fifth"] = -1
      else if(fifthModifiers == "#")
	chordData["Fifth"] = 2
      else if(fifthModifiers == "=")
	chordData["Fifth"] = -2

      while(nchar(otherModifiers) > 0)
      {
	modifiers      = sub("([0-9]+[^IViv0-9]*)(.*)", "\\1", otherModifiers)
	otherModifiers = sub("([0-9]+[^IViv0-9]*)(.*)", "\\2", otherModifiers)
	modifiedDegree = as.integer(sub("([0-9]+)([^IViv0-9]*)", "\\1", modifiers))
	modifiers      = sub("([0-9]+)([^IViv0-9]*)", "\\2", modifiers)
	if(nchar(modifiers) > 1)
	  stop("Too many modifiers in ", chord, "!")

	if(modifiedDegree > 15)
	  stop("Too high relative pitch!")

	if(modifiedDegree %in% c(9, 11, 13))
	  modifiedDegree = c("Root", NA, "Third", NA, "Fifth", "Sixth", "Seventh", NA, "Ninth", NA, "Eleventh", NA, "Thirteenth", NA, "Fifteenth")[seq(7, modifiedDegree, by=2)]
	else
	  modifiedDegree = c("Root", NA, "Third", NA, "Fifth", "Sixth", "Seventh", NA, "Ninth", NA, "Eleventh", NA, "Thirteenth", NA, "Fifteenth")[modifiedDegree]
	if(nchar(modifiers) == 0)
	  chordData[modifiedDegree] = 0
	else if(modifiers == "+")
	  chordData[modifiedDegree] = 1
	else if(modifiers == "-")
	  chordData[modifiedDegree] = -1
	else if(modifiers == "!")
	  chordData[modifiedDegree] = NA
      }
    }
  }

  if(nchar(addedNotes) > 0)
  {
    #TODO
  }

  if(nchar(pedalNote) > 0)
  {
    if(grepl("^[IViv#=\\+-]+$", pedalNote))   # pedal note specified with a roman numeral
    {
      if(is.na(tonic) || is.na(mode))
        stop("Pitch relative to tonic is specified, but no tonic is given in ", originalChord)
      relativeNote = as.integer(as.roman(pedalNote))
      if(mode == 0)
        chordData["Pedal"] = (tonic + majorDegrees[relativeNote]) %% 12
      else if(mode == 1)
        chordData["Pedal"] = (tonic + minorDegrees[relativeNote]) %% 12
      else
        stop("Incorrect mode ", mode)
      cat("Adding pedal note from", originalChord, "as", pitchNamesSharp[chordData["Pedal"]+1], "\n")
    }
    else if(grepl("^[A-Ga-g#=\\+-]+$", pedalNote))   # pedal note specified explicitly
    {
      if(any(pitchNamesSharp == pedalNote))
      {
        chordData["Pedal"] = which(pitchNamesSharp == pedalNote)[1]
        cat("Adding pedal note from", originalChord, "\n")
      }
      else if(any(pitchNamesFlat == pedalNote))
      {
        chordData["Pedal"] = which(pitchNamesFlat == pedalNote)[1]
        cat("Adding pedal note from", originalChord, "\n")
      }
      else
        stop("Unrecognized pedal point")
    }
    else
      stop("Unrecognized pedal note")
  }

  return(chordData)
}

######################################################################################################
## ksn2midi() reads a KSN file and converts it into a MIDI format
ksn2midi = function(ksnfilename, midifilename, BPM=120, basicNote=60, enharmonic=TRUE, ignoreloops=FALSE)
                                                        # 48 = C3, 60 = C4
{
  ticksPerBeat = 384

  chunkHeader = function(ID, len)  # creates a chuck header data
  {
    byteData = rep(0, 4)
    byteData[1] = floor(len/0x01000000)
    byteData[2] = floor((len - byteData[1] * 0x01000000) / 0x00010000)
    byteData[3] = floor((len - byteData[1] * 0x01000000 - byteData[2] * 0x00010000) / 0x00000100)
    byteData[4] = (len - byteData[1] * 0x01000000 - byteData[2] * 0x00010000 - byteData[3] * 0x00000100)
    return(c(charToRaw(ID), as.vector(byteData, mode="raw")))
  }

  generateVariableQuantity = function(value)
  {
    variableQuantity <- vector(0, mode="raw")
    data = rep(0, 4)
    data[1] = floor(value/(128*128*128))  # MSB
    data[2] = floor((value - data[1] * (128*128*128)) / (128*128))
    data[3] = floor((value - data[1] * (128*128) - data[2] * (128*128)) / (128))
    data[4] = (value - data[1] * (128*128*128) - data[2] * (128*128) - data[3] * (128)) # LSB
    someAlreadyWritten = FALSE
    for(byteNr in 1:4)
    {
      if((data[byteNr] > 0) || someAlreadyWritten)
      {
        variableQuantity = c(variableQuantity, as.raw(data[byteNr]))
        someAlreadyWritten = TRUE
      }
    }
    if(!someAlreadyWritten)
      variableQuantity = c(variableQuantity, as.raw(0))
    if(length(variableQuantity) > 1)
      for(byteNr in 1:(length(variableQuantity)-1))
        variableQuantity[byteNr] = as.raw(as.numeric(variableQuantity[byteNr]) + 128)  # set the MSB
    return(variableQuantity)
  }

  majorDegrees  = which(c(1,0,1,0,1,1,0,1,0,1,0,1) == 1) - 1  # a major scale
  minorDegrees  = which(c(1,0,1,1,0,1,0,1,1,0,1,0) == 1) - 1  # a harmonic minor scale
  keySignatures = rbind(c(11, 6, 1, 8, 3, 10, 5, 0, 7, 2, 9, 4, 11, 6, 1),  # major
                        c(8, 3, 10, 5, 0, 7, 2, 9, 4, 11, 6, 1, 8, 3, 10))  # minor
  notesFromHarmony = function(harmony, basicNote)
  {
    tonic = keySignatures[harmony["Mode"] + 1, harmony["Signature"] + 8]
    if(harmony["Mode"] == 0)
      degrees = majorDegrees
    else
      degrees = minorDegrees

    if(harmony["Degree"] != 0)
    {
      addDegree = function(notes, noteDegree, noteName)
	c(notes, tonic + degrees[(harmony["Degree"] + noteDegree - 2) %% 7 + 1] + harmony[noteName])
      root = tonic + degrees[harmony["Degree"]]
      if(!is.na(harmony["Root"]))
        notes = root
      else
        notes = c()
      notes = addDegree(notes, 2,  "Second")
      notes = addDegree(notes, 3,  "Third")
      notes = addDegree(notes, 4,  "Fourth")
      notes = addDegree(notes, 5,  "Fifth")
      notes = addDegree(notes, 6,  "Sixth")
      notes = addDegree(notes, 7,  "Seventh")
      notes = addDegree(notes, 9,  "Ninth")
      notes = addDegree(notes, 11, "Eleventh")
      notes = addDegree(notes, 13, "Thirteenth")
      notes = addDegree(notes, 15, "Fifteenth")

      # Remove NAs (note that are not members of this chord)
      notes = notes[!is.na(notes)]
    }
    else
      notes = c()

    # Apply chord inversion
    if((!is.na(harmony["Inversion"])) && (harmony["Inversion"] < length(notes)) && (harmony["Inversion"] > 0))
      notes = notes[c((harmony["Inversion"]+1):length(notes), 1:harmony["Inversion"])]

    # Pedal point
    if(!is.na(harmony["Pedal"]))
      notes = c(harmony["Pedal"], notes)

    # Place the notes in the right octave
    notes = notes + basicNote

    # Make sure they are in ascending pitch order
    if(length(notes) > 1)  # it might happen e.g. in Auftakt, that we only have a single note
      for(noteNr in 2:length(notes))
        if(notes[noteNr] < notes[noteNr-1])
          notes[noteNr:length(notes)] = notes[noteNr:length(notes)] + 12  # shift them up by an octave

    return(notes)
  }

  addNoteOnEvent = function(trackData, note, deltaTime=0)
  {
    trackData = c(trackData, generateVariableQuantity(deltaTime))
    return(c(trackData, as.vector(c(0x90, note, 0x7F), mode="raw"))) # note ON, channel 0
  }

  addNoteOffEvent = function(trackData, note, deltaTime=0)
  {
    trackData = c(trackData, generateVariableQuantity(deltaTime))
    return(c(trackData, as.vector(c(0x80, note, 0x7F), mode="raw"))) # note OFF, channel 0
  }

  addTempoEvent = function(trackData, BPM=120, deltaTime=0)  # tempo is beats per minute
  {
    msPerQuarterNote = 60000000 / BPM
    byteData = rep(0, 3)
    byteData[1] = floor(msPerQuarterNote / 0x00010000)
    byteData[2] = floor((msPerQuarterNote - byteData[1] * 0x00010000) / 0x00000100)
    byteData[3] = msPerQuarterNote - byteData[1] * 0x00010000 - byteData[2] * 0x00000100
    trackData = c(trackData, generateVariableQuantity(deltaTime))
    return(c(trackData, as.vector(c(0xFF, 0x51, 3, byteData), mode="raw")))
  }

  addEndOfTrackEvent = function(trackData, deltaTime=ticksPerBeat)
  {
    trackData = c(trackData, generateVariableQuantity(deltaTime)) # by default, track will end 1 beat after the last event
    return(c(trackData, as.vector(c(0xFF, 0x2F, 0x00), mode="raw")))  # End-of-track marker
  }

  addTimeSignatureEvent = function(trackData, beatsPerMeasure, ticksPerBeat, deltaTime=0)
  {
    beatValue = round(log2(96 / ticksPerBeat))
    trackData = c(trackData, generateVariableQuantity(deltaTime))
    return(c(trackData, as.vector(c(0xFF, 0x58, 4, beatsPerMeasure, beatValue, 24, 8), mode="raw")))
  }

  addKeySignatureEvent = function(trackData, keySignature, keyMode, deltaTime=0)
  {
      # these are two's complement representations of numbers -7...7
      # we have to explicitly list them because as far as I know, R does not support
      # raw _signed_ character vectors
    binarySignatures = c(0xF9, 0xFA, 0xFB, 0xFC, 0xFD, 0xFE, 0xFF, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07)
    trackData = c(trackData, generateVariableQuantity(deltaTime))
    return(c(trackData, as.vector(c(0xFF, 0x59, 2, binarySignatures[keySignature+8], keyMode), mode="raw")))
  }

  addLyricsEvent = function(trackData, lyrics, deltaTime=0)
  {
    trackData = c(trackData, generateVariableQuantity(deltaTime))
    return(c(trackData, c(as.vector(c(0xFF, 0x05, nchar(lyrics)), mode="raw"), charToRaw(lyrics))))
  }

  addTrackNameEvent = function(trackData, trackName, deltaTime=0)
  {
    trackData = c(trackData, generateVariableQuantity(deltaTime))
    return(c(trackData, c(as.vector(c(0xFF, 0x03, nchar(trackName)), mode="raw"), charToRaw(trackName))))
  }

  cat("Reading harmony...\n")
  KSData = readKSFile(ksnfilename, ignoreloops=ignoreloops)
  harmonyData = matrix(NA, nrow(KSData), 3+2+16)
  colnames(harmonyData) = c("Measures", "Beats", "Ticks", names(interpretKSKey(KSData[1, 2], enharmonic)), names(interpretKSChord(KSData[1, 3])))
  lengths = KSData[,1]
  lengths = cbind(lengths / KSData[,5], lengths)
  lengths = cbind(lengths[,1] / KSData[,4], lengths)
  harmonyData[, 1:3] = lengths
  for(ii in 1:nrow(harmonyData))
  {
    key = interpretKSKey(KSData[ii, 2], FALSE)
    harmonyData[ii, 4:ncol(harmonyData)] = c(interpretKSKey(KSData[ii, 2], enharmonic), interpretKSChord(KSData[ii, 3], tonic=key["Tonic"], mode=key["Mode"]))
  }

  cat("Generating MIDI data... ")
  timeSignature = c(harmonyData[1, "Beats"] / harmonyData[1, "Measures"], harmonyData[1, "Ticks"] / harmonyData[1, "Beats"])
  trackData = addTempoEvent(vector(length=0, mode="raw"), BPM)
  trackData = addTimeSignatureEvent(trackData, timeSignature[1], timeSignature[2])
  trackData = addKeySignatureEvent(trackData, harmonyData[1, "Signature"], harmonyData[1, "Mode"])
  trackData = addTrackNameEvent(trackData, "Harmony")
  previousEventTime = 0
  deltaTime = 0
  cumTime = c(0, cumsum(harmonyData[, "Beats"]))  # absolute time of harmony changes (in beats)
  for(harmonyNr in 1:nrow(harmonyData))
  {
    # Chord labels
    trackData = addLyricsEvent(trackData, paste(KSData[harmonyNr, "Key"], ":", KSData[harmonyNr, "Chord"], sep=""), deltaTime)

    # Time signature changes
    currentTimeSignature = c(harmonyData[harmonyNr, "Beats"] / harmonyData[harmonyNr, "Measures"], harmonyData[harmonyNr, "Ticks"] / harmonyData[harmonyNr, "Beats"])
    if(any(currentTimeSignature != timeSignature))
    {
      timeSignature = currentTimeSignature
      trackData = addTimeSignatureEvent(trackData, timeSignature[1], timeSignature[2])
    }

    # Chord notes
    notes = notesFromHarmony(harmonyData[harmonyNr,], basicNote)
    for(note in notes)
      trackData = addNoteOnEvent(trackData, note, deltaTime)
    deltaTime = (cumTime[harmonyNr+1] - previousEventTime) * ticksPerBeat
    previousEventTime = cumTime[harmonyNr+1]
    for(note in notes)
    {
      trackData = addNoteOffEvent(trackData, note, deltaTime)
      deltaTime = 0
    }
  }
  trackData = addEndOfTrackEvent(trackData)
  cat(length(trackData), " bytes \n", sep="")

  cat("Saving MIDI data...\n")
  MIDIfile = file(midifilename, "wb")
  writeBin(chunkHeader("MThd", 6), MIDIfile)
  writeBin(as.vector(c(0,0, 0,1, 0x01, 0x80), mode="raw"), MIDIfile)
    # 0,0    = format type 0 (single track MIDI)
    # 0,1    = number of tracks (1)
    # 0x8180 = 384 ticks per beat
  writeBin(chunkHeader("MTrk", length(trackData)), MIDIfile)
  writeBin(trackData, MIDIfile)
  close(MIDIfile)
}
