######################################################################################################
##
##  Reference implementation of the KSN format for GNU R
##
##  Copyright (c) 2009-2011 Stanisław A. Raczyński,
##                          Sagayama Laboratory,
##                          The University of Tokyo
##
######################################################################################################

.calculatePitchShift = function(modifiers)
{
  pitchShift = 0
  if(modifiers == "+")
    pitchShift = 1
  else if(modifiers == "-")
    pitchShift = -1
  else if(modifiers == "#")
    pitchShift = 2
  else if(modifiers == "=")
    pitchShift = 2
  return(pitchShift)
}

.getAbsolutePitchForDegree = function(scaleDegree, scaleTonic, scaleMode, pitchModifiers)
{
  if(scaleMode == 0)
    degrees = .majorDegrees
  else
    degrees = .minorNaturalDegrees

  if(scaleMode == 1 && (scaleDegree %in% c(6,7)))
    # for minor scales the sixth or the seventh can be raised
  {
    noteCandidates = scaleTonic + c(.minorNaturalDegrees[(scaleDegree - 1) %% 7 + 1], .minorMelodicDegrees[(scaleDegree - 1) %% 7 + 1]) + pitchModifiers
    return(noteCandidates[which(noteCandidates %in% .standardIntervals)])
  }
  else
    return(scaleTonic + degrees[(scaleDegree - 1) %% 7 + 1] + pitchModifiers)
}

######################################################################################################
## interpretKSKey()
##
## Parameters:
##   - key
##     A character string with key information
##   - enahrmonic
##     A Boolean indicating what format do we want the data to be returned in. If TRUE (default value),
##     it will return the number of accidentals (-7...0...+7), where negative numbers indicate
##     flats and positive numbers sharps, and a mode. If FALSE, it will return the absolute position
##     of the tonic (0..11), where 0 ic C, and a mode. Mode is zero for major keys and 1 for minor
##     ones.
##
interpretKSKey = function(key, enharmonic=TRUE)
{
  if(key %in% names(.enharmonicSpellings))
    key = .enharmonicSpellings[key]

  tonicPositions = c("C"=0, "D"=2, "E"=4, "F"=5, "G"=7, "A"=9, "B"=11)
  if(toupper(key) == key)
    mode = 0
  else
    mode = 1

  splitRegexp = "([\\+-]*)([abcdefgABCDEFG]*)"
  accidental = sub(splitRegexp, "\\1", key)
  tonic      = toupper(sub(splitRegexp, "\\2", key))
  if((nchar(tonic) != 1) || is.na(tonicPositions[tonic]))
    stop("Invalid key:", key)

  if(!enharmonic)
  {
    tonic = tonicPositions[tonic]
    if(accidental == "+")
      tonic = tonic + 1
    else if(accidental == "-")
      tonic = tonic - 1
    else if(accidental == "++")
      tonic = tonic + 2
    else if(accidental == "--")
      tonic = tonic - 2
    tonic = tonic %% 12
    names(tonic) = c()
    return(c(Tonic=tonic, Mode=mode))
  }
  else
  {
    if(mode == 0)
      Signature = which(c("-C", "-G", "-D", "-A", "-E", "-B", "F", "C", "G", "D", "A",  "E",  "B",  "+F", "+C") == key) - 8
    else
      Signature = which(c("-a", "-e", "-b", "f",  "c",  "g",  "d", "a", "e", "b", "+f", "+c", "+g", "+d", "+a") == key) - 8
    return(c(Signature=Signature, Mode=mode))
  }
}

######################################################################################################
##
## Parametrs:
##
##
## Return vector:
## - scale degree (1..7, 0 for non-chord)
## - chord type (0 = major, 1 = minor)
## - inversion (0,1,...)
## - degree modifiers: (NA, 0, +1, -1, +2, -2)
##   - root modifier
##   - 2nd modifier
##   - 3rd modifier
##   - 4th modifier
##   - 5th modifier
##   - 6th modifier
##   - 7th modifier
##   - 9th modifier
##   - 11th modifier
##   - 13th modifier
##  - added chromatic note
##  - passing chord
##
##  For note modifier: NA means that pitch is not present, 0 means that it is present.
##  Other numbers correspond to raising or lowering the pitch by the given number of semitones
##
interpretKSChord = function(chord, tonic=NA, mode=NA)
{
  chordData = c(Degree=NA, Type=NA, Inversion=NA, Root=NA, Second=NA, Third=NA, Fourth=NA, Fifth=NA, Sixth=NA, Seventh=NA, Ninth=NA, Eleventh=NA, Thirteenth=NA, Added1=NA, Added2=NA, Added3=NA, Pedal=NA, Passing=NA)

  originalChord = chord   # save the original chord for error and debug messages

  # use single characters for double flats and sharps (simpler to process)
  chord = sub("\\+\\+", "#", chord)
  chord = sub("--", "=", chord)

  # check if the chord is a passing chord, and if so, remove the parentheses
  if(grepl("^\\(.+\\)$", chord))
  {
    chordData["Passing"] = 1
    chord = sub("^\\((.+)\\)$", "\\1", chord)
  }

  # change explicit note members to added notes
  if(grepl("^\\[.+\\]", chord))
    chord = paste("q&", chord, sep="")

  chordRegexp = "^([!#=\\+-]*)([^!1-9\\=#&'/\\+-]+)([!#=\\+0-9\\-]*)('*)&?([^/&]*)/?([A-Ga-gVIvi\\+-]*)$"
  rootModifiers  = sub(chordRegexp, "\\1", chord)
  degree         = sub(chordRegexp, "\\2", chord)
  otherModifiers = sub(chordRegexp, "\\3", chord)
  inversion      = sub(chordRegexp, "\\4", chord)
  addedNotes     = sub(chordRegexp, "\\5", chord)
  pedalNote      = sub(chordRegexp, "\\6", chord)
  inversion = nchar(inversion)

  if(grepl("[^qzviVI#=\\+\\-]+", degree))
    stop("Illegal characters in the chord degree!")

  if(degree %in% c("z", "q"))
    chordData["Degree"] = 0
  else
  {
    chordData["Inversion"] = inversion
    chordData["Degree"]    = as.integer(as.roman(degree))
    chordData["Root"]      = 0
    chordData["Third"]     = 0
    chordData["Fifth"]     = 0
    if(toupper(degree) == degree)  # major chord
      chordData["Type"] = 0
    else if(tolower(degree) == degree)  # minor chord
      chordData["Type"] = 1
    else
        stop("Chord degree has mixed capital and small letters: ", chord)

    if(nchar(rootModifiers) > 1)
      stop("Too many root modifiers in ", chord, "!")
    if(rootModifiers == "!")
      chordData["Root"] = NA
    else if(rootModifiers == "+")
      chordData["Root"] = 1
    else if(rootModifiers == "-")
      chordData["Root"] = -1
    else if(rootModifiers == "#")
      chordData["Root"] = +2
    else if(rootModifiers == "=")
        chordData["Root"] = -2

    fifthModifiers = sub("([^0-9]*)([0-9]*.*)", "\\1", otherModifiers)
    otherModifiers = sub("([^0-9]*)([0-9]*.*)", "\\2", otherModifiers)
    if(nchar(fifthModifiers) > 1)
      stop("Too many fifth modifiers in ", chord, "!")
    if(fifthModifiers == "!")
      chordData["Fifth"] = NA
    else if(fifthModifiers == "+")
      chordData["Fifth"] = 1
    else if(fifthModifiers == "-")
      chordData["Fifth"] = -1
    else if(fifthModifiers == "#")
      chordData["Fifth"] = 2
    else if(fifthModifiers == "=")
      chordData["Fifth"] = -2

    while(nchar(otherModifiers) > 0)
    {
      modifiers      = sub("([0-9]+[^IViv0-9]*)(.*)", "\\1", otherModifiers)
      otherModifiers = sub("([0-9]+[^IViv0-9]*)(.*)", "\\2", otherModifiers)
      modifiedDegree = as.integer(sub("([0-9]+)([^IViv0-9]*)", "\\1", modifiers))
      modifiers      = sub("([0-9]+)([^IViv0-9]*)", "\\2", modifiers)
      if(nchar(modifiers) > 1)
        stop("Too many modifiers in ", chord, "!")

      if(modifiedDegree > 15)
        stop("Too high relative pitch!")

      if(modifiedDegree %in% c(9, 11, 13))
      {
        degreesToAdd = .degreeNames[seq(7, modifiedDegree, by=2)]
        chordData[degreesToAdd][is.na(chordData[degreesToAdd])] = 0  # add the degrees below if they are missing
      }
      modifiedDegree = .degreeNames[modifiedDegree]
      if(nchar(modifiers) == 0)
        chordData[modifiedDegree] = 0
      else if(modifiers == "+")
        chordData[modifiedDegree] = 1
      else if(modifiers == "-")
        chordData[modifiedDegree] = -1
      else if(modifiers == "!")
        chordData[modifiedDegree] = NA
    }
  }

  if(nchar(addedNotes) > 0)
  {
    if(grepl("^\\[.+\\]$", addedNotes))  # remove the brackets, if present
      addedNotes = sub("^\\[(.+)\\]$", "\\1", addedNotes)

    addedNotes = toupper(unlist(strsplit(addedNotes, " ")))
    addedNoteCount = length(addedNotes)
    if(addedNoteCount > 3)
    {
      addedNoteCount = 3
      warning("Too many added notes in ", originalChord, ". Using only the first 3.")
    }
    for(ii in 1:addedNoteCount)
    {
      elementNr = which(names(chordData) == "Added1") + ii - 1
      addedModifiers = sub("([\\+\\-#=]+)(.*)", "\\1", addedNotes[ii], perl=TRUE)
      addedNote      = sub("([\\+\\-#=]+)(.*)", "\\2", addedNotes[ii], perl=TRUE)
      addedModifiers = .calculatePitchShift(addedModifiers)
      if(grepl("^[a-gA-G]+$", addedNote))      # absolute pitch
      {
        if(any(.pitchNamesSharp == addedNote))  # we can use either .pitchNamesSharp or .pitchNamesFlat
          chordData[elementNr] = which(.pitchNamesSharp == addedNote)[1] + addedModifiers - 1
        else
          stop("Unrecognized pedal point")
        cat("Added explicit chromatic note member", addedNotes[ii], "as", .pitchNamesSharp[chordData[elementNr]+1], "in", originalChord, "\n")
      }
      else if(grepl("^[1-7]+$", addedNote))    # pitch relative to root
      {
        if(is.na(tonic))
          stop("Tonic is needed to explicitly add notes in ", originalChord)
        if(as.numeric(addedNote) > 7)
          stop("Relative added degree is too high in ", originalChord)
        chordData[which(names(chordData) == "Root") + as.numeric(addedNote) - 1] = addedModifiers
        cat("Added explicit degree", addedNotes[ii], "in", originalChord, "\n")
      }
      else if(grepl("^[IViv]+$", addedNote))   # pitch relative to the tonic
      {
        if(is.na(tonic))
          stop("Pitch relative to tonic is specified, but no tonic is given in ", originalChord)
        chordData[elementNr] = .getAbsolutePitchForDegree(as.integer(as.roman(addedNote)), tonic, mode, addedModifiers)
        chordData[elementNr] = chordData[elementNr] %% 12
        cat("Added explicit chromatic note member", addedNotes[ii], "as", .pitchNamesSharp[chordData[elementNr]+1], "in", originalChord, "\n")
      }
      else
        stop("Unrecognized explicit member note declaration in ", originalChord)
    }
  }

  if(nchar(pedalNote) > 0)
  {
    pedalModifiers = sub("([\\+\\-#=]+)(.*)", "\\1", pedalNote, perl=TRUE)
    pedalNote      = sub("([\\+\\-#=]+)(.*)", "\\2", pedalNote, perl=TRUE)
    pedalModifiers = .calculatePitchShift(pedalModifiers)
    if(grepl("^[IViv#=\\+-]+$", pedalNote))   # pedal note specified with a roman numeral
    {
      if(is.na(tonic) || is.na(mode))
        stop("Pitch relative to tonic is specified, but no tonic is given in ", originalChord)
      chordData["Pedal"] = .getAbsolutePitchForDegree(as.integer(as.roman(pedalNote)), tonic, mode, pedalModifiers)
      chordData["Pedal"] = chordData["Pedal"] %% 12
      cat("Adding pedal note from", originalChord, "as", .pitchNamesSharp[chordData["Pedal"]+1], "\n")
    }
    else if(grepl("^[A-Ga-g#=\\+-]+$", pedalNote))   # pedal note specified explicitly
    {
      if(any(.pitchNamesSharp == pedalNote))  # we can use either .pitchNamesSharp or .pitchNamesFlat
      {
        chordData["Pedal"] = which(.pitchNamesSharp == pedalNote)[1] + pedalModifiers - 1
        cat("Adding pedal note from", originalChord, "\n")
      }
      else
        stop("Unrecognized pedal point")
    }
    else
      stop("Unrecognized pedal note")
  }

  return(chordData)
}

######################################################################################################
## Returns a list of component notes for a particular chord
harmony2notes = function(harmony, basicNote=0)
{
  tonic = .keySignatures[harmony["Mode"] + 1, harmony["Signature"] + 8]

  if(harmony["Degree"] != 0)
  {
    addDegree = function(notes, noteDegree, noteName)
      c(notes, .getAbsolutePitchForDegree(harmony["Degree"] + noteDegree - 1, tonic, harmony["Mode"], harmony[noteName]))

    if(!is.na(harmony["Root"]))
      notes = .getAbsolutePitchForDegree(harmony["Degree"], tonic, harmony["Mode"], harmony["Root"])
    else
      notes = c()
    notes = addDegree(notes, 2,  "Second")
    notes = addDegree(notes, 3,  "Third")
    notes = addDegree(notes, 4,  "Fourth")
    notes = addDegree(notes, 5,  "Fifth")
    notes = addDegree(notes, 6,  "Sixth")
    notes = addDegree(notes, 7,  "Seventh")
    notes = addDegree(notes, 9,  "Ninth")
    notes = addDegree(notes, 11, "Eleventh")
    notes = addDegree(notes, 13, "Thirteenth")

    # Remove NAs (note that are not members of this chord)
    notes = notes[!is.na(notes)]
  }
  else
    notes = c()

  # Apply chord inversion
  if((!is.na(harmony["Inversion"])) && (harmony["Inversion"] < length(notes)) && (harmony["Inversion"] > 0))
    notes = notes[c((harmony["Inversion"]+1):length(notes), 1:harmony["Inversion"])]

  # Added chromatic tones
  if(!is.na(harmony["Added1"]))
  {
    warning("Ignoring added chromatic notes!")
  }

  # Pedal point
  if(!is.na(harmony["Pedal"]))
    notes = c(harmony["Pedal"], notes)

  # Place the notes in the right octave
  notes = notes + basicNote

  # Make sure they are in ascending pitch order
  if(length(notes) > 1)  # it might happen e.g. in Auftakt, that we only have a single note
    for(noteNr in 2:length(notes))
      if(notes[noteNr] < notes[noteNr-1])
        notes[noteNr:length(notes)] = notes[noteNr:length(notes)] + 12  # shift them up by an octave

  return(notes)
}

ksn2notes = function(keyLabel, chordLabel, basicNote=0)
{
  key   = interpretKSKey(keyLabel, FALSE)
  chord = interpretKSChord(chordLabel, tonic=key["Tonic"], mode=key["Mode"])
  notes = harmony2notes(c(interpretKSKey(keyLabel, TRUE), chord), basicNote=basicNote)
  return(notes)
}
