
library("KSNR")

source("dataset.R")

imageWidth  = 6  # inches
fontScaling = 0.6

splitChordLabel = function(chordLabel)
{
  if(chordLabel == "N")
    return(c("N", "N"))
  else
    unlist(strsplit(chordLabel, ":"))
}

simplifyChordType = function(chordType)
{
  if(chordType == "N")
    return(chordType)

  chordType = gsub("\\*", "", chordType)         # remove removed notes

  addedNotes = sub("^([^\\(]*)(.*)$", "\\2", chordType)
  chordType  = sub("^([^\\(]*)(.*)$", "\\1", chordType)

  if(chordType == "maj")
    chordType = "1,3,5"
  else if(chordType == "min")
    chordType = "1,b3,5"
  else if(chordType == "aug")
    chordType = "1,#3,#5"
  else if(chordType == "dim")
    chordType = "1,b3,b5"
  else if(chordType == "sus4")
    chordType = "1,b3,b5"
  else if(chordType == "sus2")
    chordType = "1,b3,b5"
  else if(chordType == "maj6")
    chordType = "1,3,5,6"
  else if(chordType == "min6")
    chordType = "1,b3,5,6"
  else if(chordType == "maj7")
    chordType = "1,3,5,7"
  else if(chordType == "min7")
    chordType = "1,b3,5,b7"
  else if(chordType == "7")
    chordType = "1,3,5,b7"
  else if(chordType == "dim7")
    chordType = "1,b3,b5,bb7"
  else if(chordType == "hdim7")
    chordType = "1,b3,b5,b7"
  else if(chordType == "minmaj7")
    chordType = "1,b3,5,7"
  else if(chordType == "9")
    chordType = "1,3,5,b7,9"
  else if(chordType == "maj9")
    chordType = "1,3,5,7,9"
  else if(chordType == "min9")
    chordType = "1,b3,5,b7,9"
  else if(chordType != "")
    stop("Unknown chord type: ", chordType)
  noteMembers = unlist(strsplit(chordType, ","))

  addedNotes = sub("^\\((.*)\\)$", "\\1", addedNotes) # remove the parentheses around the notes
  noteMembers = c(noteMembers, unlist(strsplit(addedNotes, ",")))
  noteMembers = noteMembers[sort(as.integer(gsub("[#b]", "", noteMembers)), index.return=TRUE)$ix]
     # sort by degree (omitting the sharps and flats)
  noteMembers = unique(noteMembers)  # remove repetitions

  return(paste(noteMembers, collapse=","))
}

if(exists("allData"))
  rm(allData)
for(filename in dataset)
{
  cat(filename, "\n")
  currentData = read.table(paste("../../Harmony/Harte/", filename, ".harte", sep=""), header=TRUE)
  currentData = currentData[, "Chord"]
  if(exists("allData"))
    allData = c(allData, as.character(currentData))
  else
    allData = as.character(currentData)
}

cat("Splitting chord labels...\n")
splitChords = matrix(NA, length(allData), 2)
colnames(splitChords) = c("Root", "Type")
for(chordNr in 1:length(allData))
  splitChords[chordNr,] = splitChordLabel(allData[chordNr])

cat("Simplifying chord types...\n")
for(chordNr in 1:length(allData))
  splitChords[chordNr, "Type"] = simplifyChordType(splitChords[chordNr, "Type"])

pdf("statsRoots.pdf", width=imageWidth, height=3)
par(cex=fontScaling)
rootStatistics = table(splitChords[,"Root"])
sharpNotes = c("C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B")
flatNotes  = c(NA, "Db", NA, "Eb", NA, NA, "Gb", NA, "Ab", NA, "Bb", NA)
barplot(rbind(rootStatistics[sharpNotes], rootStatistics[flatNotes]), ylab="NUmber of occurances", xlab="Pitch class")
for(ii in 1:12)
  text(-0.5 + ii*1.2, -90, flatNotes[ii], xpd=TRUE, col=rgb(0.4, 0.4, 0.4))
dev.off()

# Plot of the chord types that make up 95% of the data
chordTypeFrequencies = table(splitChords[,2])
cumSums = cumsum(sort(chordTypeFrequencies, decreasing=T)) / sum(table(splitChords[,2]))
# barplot(sort(chordTypeFrequencies, decreasing=T)[cumSums <= 0.95], las=2, ylab="Occurances")

checkChordClass = function(chordType)
{
  if(chordType == "N")
    return(c(0, 0))

  noteMembers = unlist(strsplit(chordType, ","))
  if(("#1" %in% noteMembers) || ("b1" %in% noteMembers))
    return(c(8,0))

  degreesPresent = as.integer(gsub("[#b]", "", noteMembers))
  if(max(degreesPresent) == 5 && 3 %in% degreesPresent)  # triad
  {
    if(chordType == "1,3,5")
      return(c(1,1))
    else if(chordType == "1,b3,5")
      return(c(1,2))
    else if(chordType == "1,#3,#5")
      return(c(1,3))
    else if(chordType == "1,b3,b5")
      return(c(1,4))
    else
      return(c(1,0))
  }
  else if(max(degreesPresent) == 5 && !(3 %in% degreesPresent))  # suspended
  {
    if(chordType == "1,2,3,5")
      return(c(2,1))
    else if(chordType == "1,3,4,5")
      return(c(2,2))
    else
      return(c(2,0))
  }
  else if(max(degreesPresent) == 6)  # sixth
  {
    if(chordType == "1,3,5,6")
      return(c(3,1))
    else if(chordType == "1,b3,5,6")
      return(c(3,2))
    else
      return(c(3,0))
  }
  else if(max(degreesPresent) == 7)  # seventh
  {
    if(chordType == "1,3,5,7")
      return(c(4,1))
    else if(chordType == "1,b3,5,b7")
      return(c(4,2))
    else if(chordType == "1,3,5,b7")
      return(c(4,3))
    else if(chordType == "1,b3,b5,bb7")
      return(c(4,4))
    else if(chordType == "1,b3,b5,b7")
      return(c(4,5))
    else if(chordType == "1,b3,5,7")
      return(c(4,6))
    else
      return(c(4,0))
  }
  else if(max(degreesPresent) == 9)  # ninth
  {
    if(chordType == "1,3,5,b7,9")
      return(c(5,1))
    else if(chordType == "1,3,5,7,9")
      return(c(5,2))
    else if(chordType == "1,b3,5,b7,9")
      return(c(5,3))
    else
      return(c(5,0))
  }
  else if(max(degreesPresent) == 11)  # eleventh
    return(c(6,0))
  else if(max(degreesPresent) == 13)  # thirteenth
    return(c(7,0))
  else
  {
    if(chordType == "1")
      return(c(8,1))
    else if(chordType == "1,2,3")
      return(c(8,2))
    else if(chordType == "1,3")
      return(c(8,3))
    else if(chordType == "1,b3")
      return(c(8,3))
    else if(chordType == "1,b3,4")
      return(c(8,3))
    else
    {
      cat("Unknown chord:", chordType, "\n")
      return(c(8,0))
    }
  }
}
chordTypeNames = names(chordTypeFrequencies)
chordTypes = matrix(NA, 2, length(chordTypeFrequencies))
for(ii in 1:length(chordTypeFrequencies))
  chordTypes[,ii] = checkChordClass(chordTypeNames[ii])

chordTypeMatrix = matrix(NA, 9, 7)
for(chordClassNr in 1:9) for(chordTypeNr in 1:7)
  chordTypeMatrix[chordClassNr, chordTypeNr] = sum(chordTypeFrequencies[which((chordTypes[1,] == chordClassNr - 1) & (chordTypes[2,] == chordTypeNr - 1))])
rownames(chordTypeMatrix) = c("Non-chord", "Triads", "Suspended chords", "Added sixth chords", "Seventh chords", "Ninth chords", "Eleventh chords", "Thirteenth chords", "Other")

pdf("statsChordTypes.pdf", width=imageWidth, height=7)
par(cex=fontScaling, mar=c(5, 9, 4, 4) + 0.1)
  barTops = rowSums(chordTypeMatrix)
  barColors = gray.colors(ncol(chordTypeMatrix), 0.9, 0, 1)
  plotChordTypeLegend = function(chordClassNr, typeNames)
    legend(y=-0.5 + chordClassNr*1.2, x=barTops[chordClassNr] + 300, fill=rev(barColors[1:(length(typeNames)+1)]), legend=rev(paste(c("Other", typeNames), " ", round(chordTypeMatrix[chordClassNr, 1:(length(typeNames)+1)] * 100 / barTops[chordClassNr], 1), "%", sep="")), yjust=0.5, xjust=0, density=rev(c(NA, 75)), xpd=TRUE)
  barplot(t(chordTypeMatrix), las=2, xlim=c(0, 12000), col=barColors, xlab="Number of occurances", names.arg=paste(rownames(chordTypeMatrix), "\n", round(barTops * 100 / sum(barTops)), "%", sep=""), density=c(NA, 75), horiz=TRUE)
  plotChordTypeLegend(2, c("Major", "Minor", "Diminished", "Augmented"))
  plotChordTypeLegend(4, c("Major", "Minor"))
  plotChordTypeLegend(5, c("Major seventh", "Minor seventh", "Seventh", "Diminished seventh", "Half-diminished seventh", "Minor major seventh"))
  plotChordTypeLegend(6, c("Ninth", "Major ninth", "Minor ninth"))
  plotChordTypeLegend(9, c("(1)", "(1,2,3)", "(1,3)", "(1,b3)", "(1,b3,4)"))
dev.off()
