######################################################################################################
##
##  Reference implementation of the KSN format for GNU R
##
##  Copyright (c) 2009-2011 Stanisław A. Raczyński,
##                          Sagayama Laboratory,
##                          The University of Tokyo
##
######################################################################################################

showDebugMessages = FALSE

debugmsg = function(debugMessage)
{
  if(showDebugMessages)
    cat(debugMessage, "\n")
}

#######################################################################################
## readKSFileAsTable() reads a KSN file and converts it into a table format
readKSFileAsTable = function(filename, enharmonic=TRUE, ignoreloops=FALSE)
{
  cat("Parsing the input file...\n")
  KSData = readKSFile(filename, ignoreloops=ignoreloops)
  cat("Interpreting the input file...\n")
  tableData = matrix(NA, nrow(KSData), 3+2+18)
  colnames(tableData) = c("Measures", "Beats", "Ticks", names(interpretKSKey(KSData[1, 2], enharmonic)), names(interpretKSChord(KSData[1, 3])))
  lengths = KSData[,1]
  lengths = cbind(lengths / KSData[,5], lengths)
  lengths = cbind(lengths[,1] / KSData[,4], lengths)
  tableData[, 1:3] = lengths
  for(ii in 1:nrow(tableData))
  {
    currentTonic = interpretKSKey(KSData[ii, 2], FALSE)
    if(enharmonic)
      tableData[ii, 4:ncol(tableData)] = c(interpretKSKey(KSData[ii, 2], TRUE), interpretKSChord(KSData[ii, 3], tonic=currentTonic["Tonic"], mode=currentTonic["Mode"]))
    else
      tableData[ii, 4:ncol(tableData)] = c(currentTonic, interpretKSChord(KSData[ii, 3], tonic=))
  }
  return(tableData)
}

#######################################################################################
## saveKSAsTable() saves a harmony table to a text file
saveKSAsTable = function(KSTable, filename)
  write.table(KSTable, filename, row.names=FALSE)

#######################################################################################
## readKSfile() loads a KSN file and returns a column matrix with 5 columns:
##   Ticks  - length of the chord in ticks
##   Key    - key
##   Chord  - chord
##   BPM    - number of beats per measure
##   Beat   - number of eight notes per beat, multiplied by 3 (number of ticks per beat)
##
## A tick is always a third of an eight note, so there are 96 ticks per whole note.
## To get the number of ticks per measure:
##     ticksPerMeasure = beatsPerMeasure * beatValue
## To get the meter:
##     meter = beatsPerMeasure \ (96/beatValue)
##
readKSFile = function(filename, ignoreloops=FALSE)
{
  .specialTokens = c("|", "||", ":|", "|:", ":||", "||:")

  ############################################
  ## Internal functions

  checkBraces = function(tokens, leftbrace="{", rightbrace="}", stoponerror=TRUE)
  {
    if(leftbrace == "{")
      lbracepos = grep("\\{", tokens)
    else if(leftbrace == "[")
      lbracepos = grep("\\[", tokens)
    else if(leftbrace == "(")
      lbracepos = grep("\\(", tokens)

    if(rightbrace == "}")
      rbracepos = grep("\\}", tokens)
    else if(rightbrace == "]")
      rbracepos = grep("\\]", tokens)
    else if(rightbrace == ")")
      rbracepos = grep("\\)", tokens)

    bracePositions = c(lbracepos, rbracepos)
    if(length(bracePositions) == 0)
      return(0)

    openBrackets = c(rep(1, length(lbracepos)), rep(-1, length(rbracepos)))
    sortPositions = sort(bracePositions, index.return=TRUE)
    openBrackets = openBrackets[sortPositions$ix]
    bracePositions = sortPositions$x
    openBrackets = cumsum(openBrackets)
    if(any(openBrackets < 0) && stoponerror)
    {
      errorPos = bracePositions[which(openBrackets < 0)[1]]
      if(errorPos < length(tokens))
        stop("Unexpected closing brace at: ", tokens[errorPos])
      else
        stop("Unexpected closing brace at: ", tokens[errorPos])
    }
    if((openBrackets[length(openBrackets)] != 0) && stoponerror)
      stop(paste(openBrackets[length(openBrackets)], "unmatched brackets!"))
    return(openBrackets[length(openBrackets)])
  }

  getRelativeKey = function(key, relative, isminor)
  {
    relative = sub("=[\\+-ABCDEFGabcdefg]+$", "", relative)   # remove everything after equation mark
    relative = sub("^[/0123456789]+", "", relative)
    if(grepl("^[\\+-]*[ABCDEFGabcdefg]$", relative))
      return(relative)
    keyNames = c("A", "B", "C", "D", "E", "F", "G")
    if(grepl("^[\\+-]", relative))
    {
      modifier = substr(relative, 1, 1)
      relative = substr(relative, 2, nchar(relative))
    }
    else
      modifier = NULL

    keyLetter = toupper(substr(key, nchar(key), nchar(key)))
    relativeNumber = as.integer(as.roman(relative))
    relativeKey    = keyNames[(which(keyNames == keyLetter) + relativeNumber - 2) %% 7 + 1]
    if(isminor)
      relativeKey = tolower(relativeKey)
    if(!(key %in% rownames(.keyModifiers)))
      key = .enharmonicSpellings[key]
    if(.keyModifiers[key, relativeNumber] > 0)
      relativeKey = paste("+", relativeKey, sep="")
    if(.keyModifiers[key, relativeNumber] < 0)
      relativeKey = paste("-", relativeKey, sep="")
    if(!is.null(modifier))
    {
      if(modifier == "+")
        relativeKey = paste("+", relativeKey, sep="")
      else if(modifier == "-")
        relativeKey = paste("-", relativeKey, sep="")
    }
    relativeKey = sub("\\+-", "", relativeKey)
    relativeKey = sub("-\\+", "", relativeKey)
    return(relativeKey)
  }

  compressKSsequence = function(ksSequence)
  {
    lengths   = rle(paste(ksSequence[1,], ksSequence[2,], ksSequence[3,], ksSequence[4,]))$lengths
    positions = c(1, 1+cumsum(lengths[-length(lengths)]))
    return(data.frame(Ticks=lengths, Key=ksSequence[1, positions], Chord=ksSequence[2,positions], BPM=as.numeric(ksSequence[3,positions]), Beat=as.numeric(ksSequence[4,positions]), stringsAsFactors=FALSE))
  }

  barContainsModulation = function(tokens)
  {
    result = rep(FALSE, length(tokens))
    for(ii in 1:length(tokens))
      if(regexpr("\\{[abcdefgABCDEFG\\+-iIvV]+\\:$", tokens[ii]) != -1)
        result[ii] = TRUE
    return(result)
  }

  splitLengthChord = function(chord)  # separates chord length from the rest of the chord definition
  {
    passingChord = grepl("^\\(.+\\)$", chord)
    if(passingChord)
      chord = sub("^\\((.+)\\)$", "\\1", chord)

    chordregexp  = "(^[0-9./]*)(.+$)"
    notevalue    = sub(chordregexp, "\\1", chord)
    chord        = sub(chordregexp, "\\2", chord)

    if(passingChord)
      chord = paste("(", chord, ")", sep="")

    if(notevalue == "")
      return(list(chordlen=1, chord=chord))
    else
      return(list(chordlen=notevalue, chord=chord))
  }

  isTokenChord = function(token) (regexpr("^[iIvI:+-]*[0-9/]*[abcdefgABCDEFGivIV!+-]+", token) != -1)

  modulationsLeftOpen = function(tokens)
  {
    modulationOpens  = which(barContainsModulation(tokens))
    modulationCloses = which(tokens == "}")
    for(ii in modulationCloses)
      modulationOpens[rev(which((modulationOpens < ii) & (modulationOpens != -1)))[1]] = -1
    return(tokens[modulationOpens[modulationOpens != -1]])
  }

  expandLoop = function(repBeg, repEnd, tokens)
  {
    if(is.na(repEnd) || (tokens[repEnd] == ":||"))  # normal repetition
    {
      if(is.na(repEnd))
      {
	      repEnd = length(tokens) + 1
	      beforeRep = c(tokens[1:(repBeg-1)], "|")
	      afterRep = c()
	      leftOpen = modulationsLeftOpen(beforeRep)
      }
      else if(repEnd == length(tokens))
      {
	      beforeRep = c(tokens[1:(repBeg-1)], "|")
	      afterRep = c()
	      leftOpen = modulationsLeftOpen(beforeRep)
      }
      else if(is.na(repBeg) || (repEnd < repBeg))
      {
	      repBeg = 0
	      beforeRep = c()
	      afterRep = c(tokens[(repEnd+1):length(tokens)])
	      leftOpen = c()
      }
      else
      {
	      beforeRep = c(tokens[1:(repBeg-1)], "|")
	      afterRep = c(tokens[(repEnd+1):length(tokens)])
	      if((length(afterRep) == 1) && (afterRep == ":||"))
	        afterRep = c()
	      leftOpen = modulationsLeftOpen(beforeRep)
      }
      repetition = c(tokens[(repBeg+1):(repEnd-1)], "|")
      openBraces = rep("}", checkBraces(tokens[1:(repEnd-1)], "{", "}", FALSE))
      return(c(beforeRep, repetition, openBraces, leftOpen, repetition, afterRep))
    }
    else if(tokens[repEnd] == ":|[2")  # repetition with alternative endings
    {
      if((repBeg < 1) | (repEnd > length(tokens)))
        stop("Unsupported repetition")
      firstAltBeg  = which(tokens == "|[1")
      firstAltBeg  = firstAltBeg[which((firstAltBeg > repBeg) & (firstAltBeg < repEnd))[1]]
      beforeRep    = c(tokens[1:(repBeg-1)], "|")
      repeatedPart = tokens[(repBeg+1):(firstAltBeg-1)]
      firstAlt     = tokens[(firstAltBeg+1):(repEnd-1)]
      secondAlt    = tokens[(repEnd+1):length(tokens)]
      beforeImbalance = checkBraces(beforeRep, "{", "}", FALSE)
      beforeOpen      = modulationsLeftOpen(tokens[1:(repBeg-1)])
      return(c(beforeRep, repeatedPart, "|", firstAlt, rep("}", beforeImbalance), beforeOpen, "|", repeatedPart, "|", secondAlt))
    }
    else
      stop("Internal error: bad repetition end mark")
  }

  ############################################
  ## Read the input file line by line
  lines = readLines(filename)
  tokens = c()  # list of individual tokens from the KSN file
  for(ii in 1:length(lines)) # parse the input file
  {
    curline = sub("%.*$", "", lines[ii]) # remove the comment
    curline = sub(' +$', '', curline)    # remove trailing spaces
    curline = sub('^ +', '', curline)    # remove initial spaces
    curline = sub('\\s+', ' ', curline)  # change all white space to single spaces
    tokens = c(tokens, unlist(strsplit(curline, "\\s+", perl=TRUE)))  # split into tokens on all empty characters
  }

  ############################################
  ## Pre-process the token list

  # Contract redundant and explicit chord definitions that contain spaces
  # (e.g. +vi-7-''=[III VI+ I' V']) into single tokens
  newtokens = c()
  insideChord = FALSE
  for(ii in 1:length(tokens))
  {
    if(grepl("^.+[=&]\\[", tokens[ii]) && !grepl("\\]", tokens[ii]))
    {
      insideChord = TRUE
      newtokens = c(newtokens, tokens[ii])
    }
    else if(insideChord)
    {
      if(grepl(".+\\]$", tokens[ii]))
        insideChord = FALSE
      newtokens[length(newtokens)] = paste(newtokens[length(newtokens)], tokens[ii])
    }
    else
      newtokens = c(newtokens, tokens[ii])
  }
  tokens = newtokens
  rm(newtokens)

  # Remove entries like " V}={V: I " (e.g. pivot chords).
  # Such a sequence will be converted to " } {V: I ", so only the second
  # chord interpretation will be used.
  newtokens = c()
  for(ii in 1:length(tokens))
  {
    if(grepl("^[^@]+=[^\\[]+", tokens[ii]))
    {
      closingBraces = length(unlist(strsplit(tokens[ii], "}", fixed=TRUE))) - 1
      simplifiedChord = c(rep("}", closingBraces), unlist(strsplit(tokens[ii], "=", fixed=TRUE))[2])
      newtokens = c(newtokens, simplifiedChord)
      debugmsg(paste("Simplifying redundant (e.g. pivot) chord definition", tokens[ii], " to", simplifiedChord))
    }
    else
      newtokens = c(newtokens, tokens[ii])
  }
  tokens = newtokens
  rm(newtokens)

  # Expand multi-token pedal points (e.g. "[V]&{!V9 | i'}")
  pedalPoint = ""
  for(ii in 1:length(tokens))
  {
    if(regexpr("\\[.+\\]\\&\\{.+", tokens[ii]) != -1)  # a pedal point starts here
    {
      splitToken = unlist(strsplit(tokens[ii], "\\&\\{"))
      pedalPoint = splitToken[1]
      pedalPoint = paste(splitLengthChord(unlist(strsplit(pedalPoint, "[\\[\\]]", perl=TRUE))[2])$chord, sep="")
         # ignore the length of the pedal point
      tokens[ii] = paste(splitToken[2], "/", pedalPoint, sep="")  # add pedal point at the end
      debugmsg(paste("Copying pedal point", pedalPoint))
    }
    else if((pedalPoint != "") && (regexpr(".*\\}$", tokens[ii]) != -1))  # a pedal point ends here
    {
      if(nchar(tokens[ii]) > 1)  # more than closing brace
        tokens[ii] = paste(sub("\\}", "", tokens[ii]), "/", pedalPoint, sep="")  # add pedal point at the end
      else
        tokens[ii] = ""
      pedalPoint = ""
    }
    else if((pedalPoint != "") && !(tokens[ii] %in% .specialTokens))
    {
      if(grepl("^\\(.+\\)$", tokens[ii]))
      {
        analyzedChord = sub("^\\((.+)\\)$", "\\1", tokens[ii])
        tokens[ii] = paste("(", analyzedChord, "/", pedalPoint, ")", sep="")  # add pedal point at the end
      }
      else
        tokens[ii] = paste(tokens[ii], "/", pedalPoint, sep="")  # add pedal point at the end
    }
  }
  tokens = tokens[tokens != ""]  # remove empty tokens left after removing braces that were closing pedal points

  # Split multiple closing braces and brackets joint with chord definitions
  newtokens = c()
  for(ii in 1:length(tokens))
  {
    if((bracesCount = attributes(regexpr("\\}+$", tokens[ii]))$match.length) > -1)
    {
      whatsleft = sub("\\}+$", "", tokens[ii])
      if(nchar(whatsleft) > 0)
        newtokens = c(newtokens, whatsleft, rep("}", bracesCount))
      else
        newtokens = c(newtokens, rep("}", bracesCount))
    }
    else
      newtokens = c(newtokens, tokens[ii])
  }
  tokens = newtokens
  rm(newtokens)

  # Replace repetition marks with repeated chords
  for(ii in 1:length(tokens))
  {
    if(grepl("_", tokens[ii]))
    {
      splitChord = unlist(strsplit(tokens[ii], "_"))
      if(length(splitChord) == 1)  # only "_"
      {
        leftPart = splitChord
        rightPart = ""
      }
      else
      {
        leftPart = splitChord[1]
        rightPart = splitChord[2]
      }
      for(jj in (ii-1):1)
      {
        if(isTokenChord(tokens[jj]))
        {
          middlePart = splitLengthChord(tokens[jj])$chord
          break
        }
      }
      tokens[ii] = paste(leftPart, middlePart, rightPart, sep="")
    }
  }

  # Expand higher-level loops (da capo al segno, etc.)
  while(any(tokens %in% c("@DCAF", "@DCAC", "@DSAF", "@DSAC")))
  {
    # 1. Da capo al fine
    DCAF = which(tokens == "@DCAF")
    if(length(DCAF) > 1)
      stop("There can be only one @DCAF in a song")
    else if(length(DCAF) > 0)
    {
      debugmsg("Expanding a DCAF loop")
      fine = which(tokens == "@F")
      if(length(fine) > 1)
        stop("There can be only one @F in a song")
      else if(length(fine) > 0)
      {
        imbalance = checkBraces(tokens[1:(fine-1)], "{", "}", FALSE)
        tokens = c(tokens[1:(fine-1)], tokens[(fine+1):(DCAF-1)], "|", tokens[1:(fine-1)], rep("}", imbalance), "||")
      }
      else  #TODO:Unchecked!
      {
        imbalance = checkBraces(tokens[1:(DCAF-1)], "{", "}", FALSE)
        tokens = c(tokens[1:(DCAF-1)], rep("}", imbalance), tokens[1:(DCAF-1)], tokens[(DCAF+1):length(tokens)], "||")
      }
    }

    # 2. Da capo al coda
    DCAC = which(tokens == "@DCAC")
    if(length(DCAC) > 1)
      stop("There can be only one @DCAC in a song.")
    else if(length(DCAC) > 0)
    {
      debugmsg("Expanding a DCAC loop")
      coda = which(tokens == "@C")
      if(length(coda) < 1)
        stop("@DCAC used, but no @C found.")
      else if(length(coda) > 1)
        stop("There can be only one @C in a song.")
      imbalance = checkBraces(tokens[1:(coda-1)], "{", "}", FALSE)
      tokens = c(tokens[1:(coda-1)], tokens[(coda+1):(DCAC-1)], "|", tokens[1:(coda-1)], rep("}", imbalance), "||")
    }

    DSAF = which(tokens == "@DSAF") #TODO
    if(length(DSAF) > 1)
      stop("@DSAF is not yet implemented!")

    DSAC = which(tokens == "@DSAC") #TODO
    if(length(DSAF) > 1)
      stop("@DSAC is not yet implemented!")
  }

  # Expand simple loops
  tokens[tokens == ":|"] = ":||"
  tokens[tokens == "|:"] = "||:"
  tokens[tokens == ":|:"] = ":||:"
  while(any(tokens == ":||:"))  # convert :|: to :|| ||:
  {
    loopPos = which(tokens == ":||:")[1]
    tokens = c(tokens[1:(loopPos-1)], ":||", "||:", tokens[(loopPos+1):length(tokens)])
  }
  if(!ignoreloops)
  {
    while(any((tokens == ":||") | (tokens == "||:")))  # expand all the loops
    {
      repBeg = which(tokens == "||:")[1]
      repEnd = which((tokens == ":||") | (tokens == ":|[2"))[1]
      tokens = expandLoop(repBeg, repEnd, tokens)
    }
  }
  else
  {
    tokens[tokens == ":||"] = "|"
    tokens[tokens == "||:"] = "|"
  }

  # Sanity check
  if(checkBraces(tokens, "{", "}") != 0)
    stop("Unbalanced brackets!")

  # Barmark cleanup
  tokens[tokens == "||"] = "|"      # replace "||"s with "|"s
  tokens = tokens[nchar(tokens) > 0]  # remove empty elements
  tokens = tokens[(c(FALSE, tokens[2:length(tokens)] != tokens[1:(length(tokens)-1)]) | (tokens != "|"))] # Remove repeated "|" marks

  ############################################
  ## Token analysis
  bpmSequence   = c()
  bValSequence  = c()
  chordSequence = c()
  keySequence   = c()
  curBPM   = NA
  curBVal  = NA
  curBar   = c()
  curKeys  = c()
  keyStack = c()
  for(ii in 1:length(tokens))  # process the data
  {
    curkey = keyStack[length(keyStack)]
    if((regpos = regexpr("^@M", tokens[ii])) != -1)
    {
      metrum = substr(tokens[ii], 4, nchar(tokens[ii]))
      curBPM = as.integer(sub("/.+", "", metrum))   # beats per measure
      curBVal = 96 / as.integer(sub(".+/", "", metrum))  # beat length in sixteenths
      next
    }
    if((regpos = regexpr("^@K", tokens[ii])) != -1)
    {
      keyStack = substr(tokens[ii], 4, nchar(tokens[ii]))
      curkey = keyStack[length(keyStack)]
      next
    }
    if((regpos = regexpr("^@", tokens[ii])) != -1)
    {
      stop("Unsupported directive ", tokens[ii])
      next
    }
    if((regpos = regexpr("^\\{[abcdefgABCDEFG\\+-iIvV]+\\:$", tokens[ii])) != -1)  # modulation start
    {
      newKey = substr(tokens[ii], 2, nchar(tokens[ii]) - 1)
      newKey = getRelativeKey(keyStack[length(keyStack)], newKey, newKey == tolower(newKey))
      keyStack = c(keyStack, newKey)
      curkey = newKey
      next
    }
    if((regpos = regexpr("\\}$", tokens[ii])) != -1)
    {
      keyStack = keyStack[-length(keyStack)]
      curchord = substr(tokens[ii], 1, nchar(tokens[ii]) - 1)
    }
    else
      curchord = tokens[ii]
    if(curchord == "")   # can it happen?
      next

    if(curchord != "|")
    {
      if((regexpr("^@", curchord) == -1) || (regexpr("^\\[[123456789]+$", curchord) == -1))  # ignore directives
      {
        curBar   = c(curBar, curchord)
        curKeys  = c(curKeys, curkey)
      }
    }
    else if(!is.null(curBar))
    {
      lengths = rep(0,  length(curBar))
      chords  = rep("", length(curBar))
      for(jj in 1:length(curBar))
      {
        splitchord = splitLengthChord(curBar[jj])
        if(grepl("/", splitchord$chordlen))
        {
          fraclen = strsplit(splitchord$chordlen, "/")[[1]]
          lengths[jj] = as.numeric(fraclen[1])/as.numeric(fraclen[2])
        }
        else
          lengths[jj] = as.numeric(splitchord$chordlen)
        chords[jj] = splitchord$chord
      }
      if(sum(lengths) == 0)
        lengths = 1 / length(lengths)
      else
        lengths = lengths / sum(lengths)
      lengths = lengths * curBPM * curBVal
      keys = rep("", length(chords))
      for(jj in 1:length(keys))
      {
        passingChord = grepl("^\\(.+\\)$", chords[jj])
        if(passingChord)
          analyzedChord = sub("^\\((.+)\\)$", "\\1", chords[jj])
        else
          analyzedChord = chords[jj]
        if(grepl(":", analyzedChord))
        {
          keychord = strsplit(analyzedChord, ":")[[1]]
          keys[jj] = getRelativeKey(curKeys[jj], keychord[1], keychord[1] == tolower(keychord[1]))
          if(passingChord)
            chords[jj] = paste("(", keychord[2], ")", sep="")
          else
            chords[jj] = keychord[2]
        }
        else
          keys[jj] = curKeys[jj]
      }
      for(jj in 1:length(chords))
        chords[jj] = sub("=.*$", "", chords[jj])  # remove everything after equation mark
      chordSequence = c(chordSequence, rep(chords, times=lengths))
      keySequence   = c(keySequence, rep(keys, times=lengths))
      bpmSequence   = c(bpmSequence, rep(curBPM, times=sum(lengths)))
      bValSequence  = c(bValSequence, rep(curBVal, times=sum(lengths)))
      if(length(chordSequence) != length(bpmSequence))
        stop("Looks like the tick is too small or there is a note value error in this measure: ", paste(curBar, collapse=" "))
      curBar = c()
      curKeys = c()
    }
  }

  newKeySequence = rep("", length(keySequence))
  curKey = keySequence[1]
  newKeySequence[1] = curKey
  for(ii in 2:length(keySequence))
  {
    if(keySequence[ii] != keySequence[ii-1])
      curKey = keySequence[ii]
    newKeySequence[ii] = curKey
  }

  return(compressKSsequence(rbind(keys=newKeySequence, chords=chordSequence, beatsPerMeasure=bpmSequence, beatValue=bValSequence)))
}

